/* hilbert2d.c
 * generate 2d Hilbert curves, length 2^{2,..,20}, output formatted
 * as regular '.inc' files for the Persistence Of Vision Raytracer,
 * version 3.5 or later.
 */

#define _XOPEN_SOURCE 700
#define _POSIX_C_SOURCE 200809L

#include <errno.h>
#include <locale.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>


/* conveniences + misc constants */

#define CHRDOT  '.'
#define CHRHYP  '-'
#define CHRNUL  '\0'
#define CHRPLU  '+'

#define AXIS_X  0
#define AXIS_Y  1
#define AXIS_Z  2

#define HCDIAM     0.26
#define HCLENGTH   6
#define HCSCALE    (Vector){1,1,1}
#define HCPOVVERS  3.7
#define HCPREFIX   "hc2d_"
#define HCINCEXTN  ".inc"

#define DEMOMASK  "%s_demo.pov"

#define AUXBUFSZ  512
#define AUXMASK   "// ----------<auxiliary %s '%s'>-----"

#define ZVECTOR  (Vector){0,0,0}

#define EEXIT  (eExit(__func__,__LINE__))

/* root name of curve + longest fixed 'declare'd name part */
#define HCNAMEMASK  "sweep_n%d"
#define HCNAMEMAX   22

/* longest legal POV identifier */
#define HCMAXID  40

/* scan + print */
#define VECTORMASK  "<%g, %g, %g>"


/* custom types.  a three element vector for coordinates,
 * and a sphere {<x,y,z>, radius}.
 */

typedef struct Vector Vector;
struct Vector { double x, y, z; };

typedef struct Sphere Sphere;
struct Sphere {
  union {
    Vector v;
    double xyz[3];
  } u;
  double r;
};


/* file global.  CLI stuff mostly */

static char const * const prgname = "hilbert2d";

static char const * const prgversion = "0.4.0";

/*              a auxiliary
   D demo       d diameter
                h help
   N normal     n length
                o outfile
   P pov.vers   p prefix
                r rng.seed
   S spline     s scale
   T tolerance  t translate
   V version    v verbose
                w warp */

static char const * const prgopts = "a:Dd:hN:n:o:P:p:r:S:s:T:t:Vvw:";

static char const * const prgopts_help[] = {
  "a filename\t\twith [object modifiers] to insert in data",
  "D\t\t\toutput a demo scene for the generated curve",
  "d N.N\t\tdiameter of the curve's spheres",
  "h\t\t\tdisplay this message",
  "N x | y | z\t\taxis perpendicular to curve",
  "n N\t\t\tcurve's length, two to the power of: 2,..,24",
  "o filename[.inc]\tname of the include file to write",
  "P N.N\t\trequired POV-Ray version",
  "p name\t\tcommon prefix for declared identifiers",
  "r N\t\t\tRNG sequence seed",
  "S linear | b | cubic\ttype of spline to use",
  "s <x,y,z> | N.N\tfactor(s) by which to scale curve",
  "T N.N\t\ttolerance value for spline",
  "t <x,y,z> | N.N\toffset by which to translate curve",
  "V\t\t\tdisplay the program version",
  "v\t\t\tfeedback the generator settings",
  "w N\t\t\t% limit of (random) coordinate warping",
  0
};

static bool hc_demo;

static bool hc_aux;
static char *hc_aux_value;
static FILE *hc_aux_fp;

static bool hc_diameter;
static double hc_diameter_value;

static bool hc_normal;
static int hc_normal_value;

static bool hc_length;
static int hc_length_value;

static bool hc_ofile;
static char *hc_ofile_value;

static bool hc_povvers;
static double hc_povvers_value;

static bool hc_prefix;
static char *hc_prefix_value;

static bool hc_rngseed;
static long hc_rngseed_value;

static bool hc_scale;
static Vector hc_scale_value;

static bool hc_spline;
static size_t hc_spline_value;

static bool hc_tolerance;
static double hc_tolerance_value;

static bool hc_translate;
static Vector hc_translate_value;

static bool hc_verbose;

static bool hc_warp;
static int hc_warp_value;

static Vector hc_curve_begin,
              hc_curve_end;

static char *hc_root_name;

/* nb  code relies on their order */
static char const * const hc_splinetypes[] = {
  "linear_spline", "b_spline", "cubic_spline", 0
};


/* -------------------------------------------------------------------------- *
 * oops + GNU mandated interface.
 */

static void eExit(char const * const, int const) __attribute__((noreturn));
static void eExit(char const * const fn, int const ln) {
  int const e = errno;
  fprintf(stderr, "%s: oops in '%s' line %d, error (%d) %s.\n",
          prgname, fn, ln, e, strerror(e));
  exit(EXIT_FAILURE);
}

static void usageHelp(int const) __attribute__((noreturn));
static void usageHelp(int const ec) {
  char const * const fmt = "%s - Hilbert curve POV include file generator.\n"
          "\n  usage: %s [options]\n"
          "\nthe available options are:\n";
  printf(fmt, prgname, prgname);
  for (size_t i = 0; prgopts_help[i]; ++i)
      printf("  -%s.\n", prgopts_help[i]);
  printf("\nsee \"man 1 %s\" for details.\n", prgname);
  exit(ec);
}

static void usageVersion(void) __attribute__((noreturn));
static void usageVersion(void) {
  char const * const fmt = "%s %s\n"
          "copyright (c) 2014-2018 jr <creature.eternal@gmail.com>.\n"
          "license GPLv3+ <http://gnu.org/licenses/gpl.html>\n"
          "this is free software, you are free to .. whatever.\n"
          "there is no warranty.\n";
  printf(fmt, prgname, prgversion);
  exit(EXIT_SUCCESS);
}


/* -------------------------------------------------------------------------- *
 * code courtesy of en.wikipedia.org/wiki/Hilbert_curve.
 */

static inline void rotFlip(int n, int *x, int *y, int rx, int ry) {
  if (0 == ry) {
    if (1 == rx) {
      *x = n - 1 - *x;
      *y = n - 1 - *y;
    }
    int t = *x;
    *x = *y;
    *y = t;
  }
}

static void d2xy(int n, int d, int *x, int *y) {
  int t = d;
  int rx, ry;
  *x = *y = 0;
  for (int s = 1; s < n; s *= 2) {
    rx = 1 & (t / 2);
    ry = 1 & (t ^ rx);
    rotFlip(s, x, y, rx, ry);
    *x += s * rx;
    *y += s * ry;
    t /= 4;
  }
}


/* -------------------------------------------------------------------------- *
 * helpers.  no (overflow) checks.
 */

static inline Vector vectAdd(Vector const v1, Vector const v2) {
  return (Vector){v1.x + v2.x, v1.y + v2.y, v1.z + v2.z};
}

static inline Vector vectMult(Vector const v1, Vector const v2) {
  return (Vector){v1.x * v2.x, v1.y * v2.y, v1.z * v2.z};
}

static inline Vector vectWarp(Vector const v) {
  if (!hc_warp) return v;
  /* warp limit */
  double const wl = (double)hc_warp_value / 100.;
  Sphere s;
  s.u.v = v;
  for (size_t i = 0; 3 > i; ++i) {
    double n;
label_again:
    n = drand48();
    if (wl < n)
      goto label_again;
    if ((int)(100000 * n) % 2)
      s.u.xyz[i] -= n;
    else
      s.u.xyz[i] += n;
  }
  return s.u.v;
}


/* -------------------------------------------------------------------------- *
 * outputting stuff.
 */

static void genVerbose(void) {
  int const w = 18;
  printf("generator settings ('+' argument, '-' default):\n");
  printf("%*s: %c %g\n", w, "povray version", hc_povvers ? CHRPLU : CHRHYP,
          hc_povvers_value);
  printf("%*s: %c '%s'\n", w, "prefix", hc_prefix ? CHRPLU : CHRHYP,
          hc_prefix_value ? hc_prefix_value : HCPREFIX);
  printf("%*s: %c %g\n", w, "diameter", hc_diameter ? CHRPLU : CHRHYP,
          hc_diameter_value);
  printf("%*s: %c '%c' axis\n", w, "normal", hc_normal ? CHRPLU : CHRHYP,
          120 + hc_normal_value);
  printf("%*s: %c " VECTORMASK "\n", w, "scale", hc_scale ? CHRPLU : CHRHYP,
          hc_scale_value.x, hc_scale_value.y, hc_scale_value.z);
  printf("%*s: %c " VECTORMASK "\n", w, "translate",
          hc_translate ? CHRPLU : CHRHYP,
          hc_translate_value.x, hc_translate_value.y, hc_translate_value.z);
  printf("%*s: %c %d %sspheres\n", w, "curve length",
          hc_length ? CHRPLU : CHRHYP, (int)pow(2, hc_length_value),
          hc_spline_value ? "(+2) " : "");
  printf("%*s: %c '%s'\n", w, "spline type", hc_spline ? CHRPLU : CHRHYP,
          hc_splinetypes[hc_spline_value]);
  if (hc_tolerance)
    printf("%*s: %c %g\n", w, "tolerance", hc_tolerance ? CHRPLU : CHRHYP,
            hc_tolerance_value);
  if (hc_aux)
    printf("%*s: %c '%s'\n", w, "auxiliary", CHRPLU, hc_aux_value);
  printf("%*s: %c %d%%\n", w, "warp limit", hc_warp ? CHRPLU : CHRHYP,
          hc_warp_value);
  printf("%*s: %c %ld\n", w, "rng seed", hc_rngseed ? CHRPLU : CHRHYP,
          hc_rngseed_value);
  printf("%*s: %c '%s'\n", w, "out file", hc_ofile ? CHRPLU : CHRHYP,
          hc_ofile_value ? hc_ofile_value : "stdout");
  if (hc_ofile)
    printf("%*s: %c\n", w, "demo", hc_demo ? CHRPLU : CHRHYP);
  printf("%*s:   " VECTORMASK ", " VECTORMASK "\n", w, "begin,end points",
          hc_curve_begin.x, hc_curve_begin.y, hc_curve_begin.z,
          hc_curve_end.x, hc_curve_end.y, hc_curve_end.z);
  return;
}

static void genDemo(char const * const fn) {
  FILE *fp;
  if (!(fp = fopen(fn, "wb"))) EEXIT;
  char const * const fmt = "// Hilbert 2D curve demo/test scene\n"
          "//   +a0.3 +w800 +h600\n"
          "#version %g;\n"
          "#include \"colors.inc\"\n"
          "global_settings { assumed_gamma 1.0 }\n"
          "light_source { " VECTORMASK " White }\n"
          "camera {\n"
          "  location " VECTORMASK "\n"
          "  look_at " VECTORMASK "\n"
          "  right x * (image_width/image_height)\n"
          "}\n"
          "#declare %s%s_arg_diameter = %g;\n"
          "#include \"%s\"\n"
          "plane { y, -1.5 pigment { Gray25 } }\n"
          "object { %s%s no_shadow pigment { Red } }\n"
          "// end of demo/test scene\n";
  Vector const light = vectAdd(hc_translate_value, (Vector){0., 100., 0.});
  /* TODO camera cf curve length, scale, normal */
  Vector const cp = vectAdd(hc_translate_value, (Vector){0., 4., -4.});
  Vector const cl = vectAdd(hc_translate_value, (Vector){3., 1., 3.});
  fprintf(fp, fmt, hc_povvers_value,
          light.x, light.y, light.z,
          cp.x, cp.y, cp.z,
          cl.x, cl.y, cl.z,
          hc_prefix_value, hc_root_name, hc_diameter_value,
          hc_ofile_value,
          hc_prefix_value, hc_root_name);
  if (fclose(fp)) EEXIT;
  return;
}

static void genPreamble(FILE *fp) {
  char const * const fmt = "// Hilbert 2D curve\n"
          "// generated %s (%s v %s).\n\n"
          "#ifndef(%s%s_include_temp)\n\n"
          "#declare %s%s_include_temp = version;\n"
          "#version %g;\n\n"
          "#ifdef(View_POV_Include_Stack)\n"
          "#debug \"including '%s%s'\\n\"\n"
          "#end\n\n"
          "#ifdef(%s%s_arg_diameter)\n"
          "#local HCR = %s%s_arg_diameter / 2;\n"
          "#else\n"
          "#local HCR = %g;\n"
          "#end\n\n";
  time_t const now = time(0);
  struct tm *tm = localtime(&now);
  char buf[41] = {CHRNUL};
  strftime(buf, 40, "%d.%m.%Y %H%Mh", tm);
  fprintf(fp, fmt, buf, prgname, prgversion,
          hc_prefix_value, hc_root_name,
          hc_prefix_value, hc_root_name,
          hc_povvers_value,
          hc_ofile_value ? "" : hc_prefix_value,
          hc_ofile_value ? hc_ofile_value : hc_root_name,
          hc_prefix_value, hc_root_name,
          hc_prefix_value, hc_root_name,
          hc_diameter_value / 2);
  if (ferror(fp)) EEXIT;
  return;
}

static void genPostamble(FILE *fp) {
  char const * const fmt = "\n};\n\n"
          "#declare %s%s_begin = " VECTORMASK ";\n"
          "#declare %s%s_end = " VECTORMASK ";\n\n"
          "#version %s%s_include_temp;\n\n"
          "#end\n";
  fprintf(fp, fmt,
          hc_prefix_value, hc_root_name,
          hc_curve_begin.x, hc_curve_begin.y, hc_curve_begin.z,
          hc_prefix_value, hc_root_name,
          hc_curve_end.x, hc_curve_end.y, hc_curve_end.z,
          hc_prefix_value, hc_root_name);
  if (ferror(fp)) EEXIT;
  return;
}

static void genAuxiliary(FILE *fp) {
  fprintf(fp, "\n" AUXMASK "\n", "begin", hc_aux_value);
  char buf[AUXBUFSZ] = {CHRNUL};
  while (fgets(buf, AUXBUFSZ, hc_aux_fp))
    if (EOF == fputs(buf, fp)) EEXIT;
  if (ferror(hc_aux_fp)) EEXIT;
  fprintf(fp, AUXMASK, "end", hc_aux_value);
  if (ferror(fp)) EEXIT;
  return;
}

static void genCurve(FILE *fp) {
  char const * const fmt = "#declare %s%s = sphere_sweep {\n  %s\n  %d";
  int const N = (int)pow(2, hc_length_value);
  double const offs = 0.0001;
  /* b + cubic spline have leading/trailing 'control's */
  if (hc_spline_value) {
    Vector c = ZVECTOR;
    switch (hc_normal_value) {
      case AXIS_X:  c.y = c.z = -offs;  break;
      case AXIS_Y:  c.x = c.z = -offs;  break;
      case AXIS_Z:  c.x = c.y = -offs;  break;
    }
    Vector v = vectAdd(hc_translate_value, c);
    fprintf(fp, fmt, hc_prefix_value, hc_root_name,
            hc_splinetypes[hc_spline_value], N + 2);
    fprintf(fp, ",\n  " VECTORMASK ", HCR", v.x, v.y, v.z);
  }
  else
    fprintf(fp, fmt, hc_prefix_value, hc_root_name,
            hc_splinetypes[hc_spline_value], N);
  int x, y;
  for (int i = x = y = 0; i < N; ++i) {
    Vector v;
    d2xy(N, i, &x, &y);
    switch (hc_normal_value) {
      case AXIS_X:  v = vectAdd((Vector){0, x, y}, ZVECTOR);  break;
      case AXIS_Y:  v = vectAdd((Vector){x, 0, y}, ZVECTOR);  break;
      case AXIS_Z:  v = vectAdd((Vector){x, y, 0}, ZVECTOR);  break;
    }
    v = vectWarp(v);
    v = vectMult(hc_scale_value, v);
    v = vectAdd(hc_translate_value, v);
    fprintf(fp, ",\n  " VECTORMASK ", HCR", v.x, v.y, v.z);
    if (ferror(fp)) EEXIT;
    if (i == (N - 1))
      hc_curve_end = v;
    else if (!i)
      hc_curve_begin = v;
  }
  if (hc_spline_value) {
    Vector c = ZVECTOR;
    switch (hc_normal_value) {
      case AXIS_X:  c.y = c.z = offs;  break;
      case AXIS_Y:  c.x = c.z = offs;  break;
      case AXIS_Z:  c.x = c.y = offs;  break;
    }
    Vector v = vectAdd(hc_curve_end, c);
    fprintf(fp, ",\n  " VECTORMASK ", HCR", v.x, v.y, v.z);
  }
  if (hc_tolerance) fprintf(fp, "\n  tolerance %g", hc_tolerance_value);
  if (ferror(fp)) EEXIT;
  return;
}


/* -------------------------------------------------------------------------- *
 * generator controls data creation and output.
 */

static int generator(void) {
  /* .inc == preamble + curve + aux + postamble */
  FILE *fp;
  if (!hc_ofile) fp = stdout;
  else if (!(fp = fopen(hc_ofile_value, "wb"))) EEXIT;
  genPreamble(fp);
  genCurve(fp);
  if (hc_aux) genAuxiliary(fp);
  genPostamble(fp);
  if (hc_ofile && fclose(fp)) EEXIT;
  /* demo requires stored include */
  if (hc_ofile && hc_demo) {
    char buf[strlen(hc_ofile_value) + strlen(DEMOMASK)];
    (void)sprintf(buf, DEMOMASK, hc_ofile_value);
    genDemo(buf);
  }
  /* feedback always to stdout */
  if (hc_verbose) genVerbose();
  if (hc_ofile) printf("%swrote '%s'%s.\n",
          hc_verbose ? "\n" : "", hc_ofile_value,
          hc_demo ? ", and demo scene" : "");
  return EXIT_SUCCESS;
}


/* -------------------------------------------------------------------------- *
 * specific function ('argXxx') for each option with an argument,
 * using the 'getXxx' functions as needed.  true when successful.
 */

static bool getDouble(char const *arg, double *val) {
  errno = 0;
  char *p = 0;
  double d = strtod(arg, &p);
  if (errno || *p) return false;
  *val = d;
  return true;
}

static bool getLong(char const *arg, long *val) {
  errno = 0;
  char *p = 0;
  long n = strtol(arg, &p, 10);
  if (errno || *p) return false;
  *val = n;
  return true;
}

static bool getVector(char const *arg, Vector *val) {
  float x, y, z;
  if (3 != sscanf(arg, VECTORMASK, &x, &y, &z))
    return false;
  *val = (Vector){x, y, z};
  return true;
}

static bool argAuxiliary(char const *arg) {
  hc_aux_fp = fopen(arg, "rb");
  if (!hc_aux_fp) EEXIT;
  hc_aux_value = strdup(arg);
  if (!hc_aux_value) {
    (void)fclose(hc_aux_fp);
    EEXIT;
  }
  hc_aux = true;
  return true;
}

static bool argDiameter(char const *arg) {
  double d = 0;
  if (!getDouble(arg, &d))
    return false;
  else if (0.000001 > d  ||  2.5 <= d)
    return false;
  hc_diameter_value = d;
  hc_diameter = true;
  return true;
}

static bool argNormal(char const *arg) {
  bool flag = true;
  hc_normal = true;
  switch (*arg) {
    case 'x':  hc_normal_value = AXIS_X;  break;
    case 'y':  hc_normal_value = AXIS_Y;  break;
    case 'z':  hc_normal_value = AXIS_Z;  break;
    default:  flag = false;
  }
  return flag;
}

static bool argLength(char const *arg) {
  long n = 0;
  if (!getLong(arg, &n))
    return false;
  else if (2 > n  ||  24 < n)
    return false;
  hc_length_value = (int)n;
  hc_length = true;
  return true;
}

static bool argOutfile(char const *arg) {
  char const * const extn = strrchr(arg, CHRDOT);
  size_t const n = strlen(arg);
  /* length check against (longer) demo name */
  if (!n)
    return false;
  else if (pathconf(".", _PC_PATH_MAX) <= (long)(n + extn ? 11 : 15))
    return false;
  char *fn = 0;
  if (extn)
    fn = strdup(arg);
  else {
    char buf[n + 5];
    (void)sprintf(buf, "%s%s", arg, HCINCEXTN);
    fn = strdup(buf);
  }
  if (!fn) EEXIT;
  FILE *fp = fopen(fn, "r");  /* no overwrite */
  if (fp) {
    (void)fclose(fp);
    errno = EEXIST;
    EEXIT;
  }
  hc_ofile_value = fn;
  hc_ofile = true;
  return true;
}

static bool argPovvers(char const *arg) {
  double d = 0;
  if (!getDouble(arg, &d))
    return false;
  else if (3.5 > d  ||  4.0 <= d)
    return false;
  hc_povvers_value = d;
  hc_povvers = true;
  return true;
}

static bool argPrefix(char const *arg) {
  size_t const n = strlen(arg);
  if (!n || HCMAXID <= (n + HCNAMEMAX))
    return false;
  /* no verification of content */
  hc_prefix_value = strdup(arg);
  if (!hc_prefix_value) EEXIT;
  hc_prefix = true;
  return true;
}

static bool argRngseed(char const *arg) {
  long n = 0;
  if (!getLong(arg, &n))
    return false;
  else if (!n)
    return false;
  hc_rngseed_value = labs(n);
  hc_rngseed = true;
  return true;
}

static bool argScale(char const *arg) {
  double d = 0;
  Vector v = ZVECTOR;
  if (getDouble(arg, &d))
    v = (Vector){d, d, d};
  else if (!getVector(arg, &v))
    return false;
  if (!(v.x && v.y && v.z))
    return false;
  hc_scale_value = v;
  hc_scale = true;
  return true;
}

static bool argSpline(char const *arg) {
  for (size_t i = 0; hc_splinetypes[i]; ++i)
    if (hc_splinetypes[i][0] == *arg) {
      hc_spline_value = i;
      hc_spline = true;
      return true;
    }
  return false;
}

static bool argTolerance(char const *arg) {
  double d = 0;
  if (!getDouble(arg, &d))
    return false;
  else if (0.000001 > d  ||  1.0 <= d)
    return false;
  hc_tolerance_value = d;
  hc_tolerance = true;
  return true;
}

static bool argTranslate(char const *arg) {
  double d = 0;
  Vector v = ZVECTOR;
  if (getDouble(arg, &d))
    v = (Vector){d, d, d};
  else if (!getVector(arg, &v))
    return false;
  hc_translate_value = v;
  hc_translate = true;
  return true;
}

static bool argWarp(char const *arg) {
  long n = 0;
  if (!getLong(arg, &n))
    return false;
  else if (1 > n  ||  100 < n)
    return false;
  hc_warp_value = (int)n;
  hc_warp = true;
  return true;
}


/* -------------------------------------------------------------------------- *
 * "sensible" defaults.  zeroing + bools to false is redundant, but hey..
 */

static void hcDefaults(void) {
  hc_aux = false, hc_aux_value = 0;
  hc_demo = false;
  hc_diameter = false, hc_diameter_value = HCDIAM;
  hc_warp = false, hc_warp_value = 0;
  hc_normal = false, hc_normal_value = AXIS_Y;
  hc_length = false, hc_length_value = HCLENGTH;
  hc_ofile = false, hc_ofile_value = 0;
  hc_povvers = false, hc_povvers_value = HCPOVVERS;
  hc_prefix = false, hc_prefix_value = 0;
  hc_rngseed = false, hc_rngseed_value = (long)time(0);
  hc_scale = false, hc_scale_value = HCSCALE;
  hc_spline = false, hc_spline_value = 0;
  hc_tolerance = false, hc_tolerance_value = 0;
  hc_translate = false, hc_translate_value = ZVECTOR;
  hc_curve_begin = hc_curve_end = ZVECTOR;
  return;
}


/* -------------------------------------------------------------------------- *
 * main's job, initialisation aside, is to deal with the command-line and,
 * when done, return via the generator function.
 */

static void tidyup(void) {
  if (hc_aux) (void)fclose(hc_aux_fp);
  free(hc_aux_value);
  free(hc_ofile_value);
  free(hc_prefix_value);
  free(hc_root_name);
  return;
}

int main(int argc, char *argv[]) {
  if (atexit(tidyup)) EEXIT;
  else if (setreuid(getuid(), geteuid())) EEXIT;
  else if (!setlocale(LC_ALL, "")) EEXIT;
  hcDefaults();
  opterr = 0;
  int o;
  while (-1 != (o = getopt(argc, argv, prgopts)))
    switch (o) {
      case 'a':  /* always successful (if returns) */
        if (hc_aux || !argAuxiliary(optarg)) {
          fprintf(stderr, "%s: oops, expected auxiliary file, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'D':  hc_demo = true;  break;
      case 'd':
        if (hc_diameter || !argDiameter(optarg)) {
          fprintf(stderr,
                  "%s: oops, expected \"diameter\", got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'h':  usageHelp(0);
      case 'N':
        if (hc_normal || !argNormal(optarg)) {
          fprintf(stderr, "%s: oops, expected \"normal\" axis, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'n':
        if (hc_length || !argLength(optarg)) {
          fprintf(stderr, "%s: oops, expected \"length\" 2,..20, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'o':
        if (hc_ofile || !argOutfile(optarg)) {
          fprintf(stderr, "%s: oops, expected include file name, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'P':
        if (hc_povvers || !argPovvers(optarg)) {
          fprintf(stderr,
                  "%s: oops, expected \"version\" 3.5+, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'p':
        if (hc_prefix || !argPrefix(optarg)) {
          fprintf(stderr, "%s: oops, unexpected (length) identifier prefix,"
                  " got '%s'.\n", prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'r':
        if (hc_rngseed || !argRngseed(optarg)) {
          fprintf(stderr, "%s: oops, expected rng seed, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'S':
        if (hc_spline || !argSpline(optarg)) {
          fprintf(stderr, "%s: oops, expected spline type, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 's':
        if (hc_scale || !argScale(optarg)) {
          fprintf(stderr,
                  "%s: oops, expected scale vector or float, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'T':
        if (hc_tolerance || !argTolerance(optarg)) {
          fprintf(stderr,
                  "%s: oops, expected \"tolerance\" for spline, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 't':
        if (hc_translate || !argTranslate(optarg)) {
          fprintf(stderr,
                  "%s: oops, expected coordinate vector or float, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      case 'V':  usageVersion();
      case 'v':  hc_verbose = true;  break;
      case 'w':
        if (hc_warp || !argWarp(optarg)) {
          fprintf(stderr, "%s: oops, expected \"warp\" limit, got '%s'.\n",
                  prgname, optarg);
          exit(EXIT_FAILURE);
        }
        break;
      default:   usageHelp(2);
    }

  if (argc != optind) {
    fprintf(stderr, "%s: oops, unknown argument(s) on command line.\n",
            prgname);
    exit(EXIT_FAILURE);
  }
  /* final preps, incl length-based root name */
  {
    char buf[HCMAXID] = {CHRNUL};
    (void)sprintf(buf, HCNAMEMASK, hc_length_value);
    if (!(hc_root_name = strdup(buf))) EEXIT;
  }
  if (!hc_prefix) if (!(hc_prefix_value = strdup(HCPREFIX))) EEXIT;
  srand48(hc_rngseed_value);

  return generator();
}

/* ------------------------------------------------------------- *
 * the content above is covered by the GPLv3+, see file COPYING. *
 * copyright (c) 2018 jr <creature.eternal@gmail.com>.           *
 * all rights reserved.                                          *
 * ------------------------------------------------------------- *
 */
